#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility


if __name__ == "__main__":
    do_debug = False
    paths = utility.get_default_test_paths(
        __file__, "test050_let_actor_letd", "test050"
    )
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 1234567891
    sim.number_of_threads = 2
    sim.output_dir = paths.output

    numPartSimTest = 40000 / sim.number_of_threads
    numPartSimRef = 1e5

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    # world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 10 * cm, 10 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    test_material_name = "G4_WATER"
    phantom_off = sim.add_volume("Box", "phantom_off")
    phantom_off.mother = phantom.name
    phantom_off.size = [100 * mm, 60 * mm, 60 * mm]
    phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
    phantom_off.material = test_material_name
    phantom_off.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"
    # sim.physics_manager.set_production_cut("world", "all", 1000 * km)
    # FIXME need SetMaxStepSizeInRegion ActivateStepLimiter
    # now available
    # e.g.
    # sim.physics_manager.set_max_step_size(volume_name='phantom', max_step_size=1*mm)

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 80 * MeV
    # source.energy.type = 'gauss'
    # source.energy.sigma_gauss = 1 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.radius = 4 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    # print(dir(source.energy))
    source.n = numPartSimTest
    # source.activity = 100 * kBq

    # filter : keep proton
    # f = sim.add_filter("ParticleFilter", "f")
    # f.particle = "proton"

    size = [50, 1, 1]
    spacing = [2.0 * mm, 60.0 * mm, 60.0 * mm]

    doseActorName_IDD_d = "IDD_d"
    doseIDD = sim.add_actor("DoseActor", doseActorName_IDD_d)
    doseIDD.output_filename = "test050-" + doseActorName_IDD_d + ".mhd"
    doseIDD.attached_to = phantom_off
    doseIDD.size = size
    doseIDD.spacing = spacing
    doseIDD.hit_type = "random"
    doseIDD.dose.active = False

    LETActorName_IDD_d = "LETActorOG_d"
    LETActor_IDD_d = sim.add_actor("LETActor", LETActorName_IDD_d)
    LETActor_IDD_d.output_filename = "test050-" + LETActorName_IDD_d + ".mhd"
    LETActor_IDD_d.attached_to = phantom_off
    LETActor_IDD_d.size = size
    LETActor_IDD_d.spacing = spacing
    LETActor_IDD_d.hit_type = "random"
    LETActor_IDD_d.averaging_method = "dose_average"

    LETActorName_IDD_t = "LETActorOG_t"
    LETActor_IDD_t = sim.add_actor("LETActor", LETActorName_IDD_t)
    LETActor_IDD_t.output_filename = "test050-" + LETActorName_IDD_t + ".mhd"
    LETActor_IDD_t.attached_to = phantom_off
    LETActor_IDD_t.size = size
    LETActor_IDD_t.spacing = spacing
    LETActor_IDD_t.hit_type = "random"
    LETActor_IDD_t.averaging_method = "track_average"

    LETActorName_IDD_d2w = "LETActorOG_d2w"
    LETActor_IDD_d2w = sim.add_actor("LETActor", LETActorName_IDD_d2w)
    LETActor_IDD_d2w.output_filename = "test050-" + LETActorName_IDD_d2w + ".mhd"
    LETActor_IDD_d2w.attached_to = phantom_off
    LETActor_IDD_d2w.size = size
    LETActor_IDD_d2w.spacing = spacing
    LETActor_IDD_d2w.hit_type = "random"
    LETActor_IDD_d2w.score_in = "G4_WATER"
    LETActor_IDD_d2w.averaging_method = "dose_average"

    LET_primaries = "LETprimaries"
    LETActor_primaries = sim.add_actor("LETActor", LET_primaries)
    LETActor_primaries.output_filename = "test050-" + LET_primaries + ".mhd"
    LETActor_primaries.attached_to = phantom_off
    LETActor_primaries.size = size
    LETActor_primaries.spacing = spacing
    LETActor_primaries.hit_type = "random"
    LETActor_primaries.averaging_method = "dose_average"

    # # add dose actor, without e- (to check)
    fe = sim.add_filter("ParticleFilter", "f")
    fe.particle = "proton"
    fe.policy = "accept"
    LETActor_primaries.filters.append(fe)
    print(fe)

    fName_ref_IDD = "IDD__Proton_Energy1MeVu_RiFiout-Edep.mhd"
    print(paths)
    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True
    # stats.filters.append(f)

    print("Filters: ", sim.filter_manager)
    # print(sim.filter_manager.dump())

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    print()
    # gate.exception.warning("Tests stats file")
    # stats_ref = utility.read_stat_file(paths.gate_output / "stats.txt")
    # is_ok = utility.assert_stats(stat, stats_ref, 0.14)

    fNameIDD = doseIDD.edep.output_filename
    if do_debug:
        is_ok = utility.assert_images(
            ref_path / fNameIDD,
            doseIDD.edep.get_output_path(),
            stats,
            tolerance=100,
            ignore_value_data2=0,
            axis="x",
            scaleImageValuesFactor=numPartSimRef / numPartSimTest,
        )

    tests_pass = []
    is_ok = utility.assert_filtered_imagesprofile1D(
        ref_filter_filename1=str(doseIDD.edep.get_output_path()),
        ref_filename1=ref_path
        / "test050_LET1D_noFilter__PrimaryProton-doseAveraged.mhd",
        filename2=str(LETActor_IDD_d.let.get_output_path()),
        tolerance=60,
        # plt_ylim=[0, 25],
    )
    tests_pass.append(is_ok)
    print(f"218 {is_ok =}")

    is_ok = (
        utility.assert_filtered_imagesprofile1D(
            ref_filter_filename1=str(doseIDD.edep.get_output_path()),
            ref_filename1=ref_path
            / "test050_LET1D_noFilter__PrimaryProton-trackAveraged.mhd",
            filename2=LETActor_IDD_t.let.get_output_path(),
            tolerance=8,
            # plt_ylim=[0, 18],
        )
        and is_ok
    )
    is_ok = (
        utility.assert_filtered_imagesprofile1D(
            ref_filter_filename1=str(doseIDD.edep.get_output_path()),
            ref_filename1=ref_path / "test050_LET1D_Z1__PrimaryProton-doseAveraged.mhd",
            filename2=LETActor_primaries.let.get_output_path(),
            tolerance=5,
            # plt_ylim=[0, 25],
        )
        and is_ok
    )
    print(f"222 {is_ok =}")
    tests_pass.append(is_ok)

    utility.test_ok(all(tests_pass))



from box import Box
from functools import wraps

from ..definitions import __world_name__
from ..exception import fatal, GateImplementationError
from ..base import GateObject, process_cls
from ..utility import insert_suffix_before_extension
from .actoroutput import ActorOutputRoot


def _setter_hook_attached_to(self, attached_to):
    """Hook to be attached to property setter of user input 'attached_to' in all actors.
    Allows the user input 'attached_to_volume' to be volume object or a volume name.
    """

    if isinstance(attached_to, str):
        return attached_to
    else:
        try:
            return attached_to.name
        except AttributeError:
            if isinstance(
                attached_to,
                (
                    list,
                    tuple,
                ),
            ):
                attached_to_names = []
                for a in attached_to:
                    try:
                        attached_to_names.append(a.name)
                    except AttributeError:
                        attached_to_names.append(a)
                return attached_to_names

    # something went wrong if we reached this point
    fatal(
        f"attached_to must be a volume, a volume name, "
        f"or a list/tuple of volumes or volume names. "
        f"Received: {attached_to}"
    )


def shortcut_for_single_output_actor(func):
    """Decorator for shortcut methods and properties that may be used only
    with actors that handle a single user output."""

    @wraps(func)
    def _with_check(self, *args):
        if len(self.interfaces_to_user_output) > 1:
            try:
                # func could be a method,
                name = func.__name__
            except AttributeError:
                try:
                    # or a property
                    name = func.fget.__name__
                except AttributeError:
                    name = ""
            s = (
                f"The shortcut {name} is not available for actor {self.type_name} "
                f"because the actor handles more than one interface to output, "
                f"namely {list(self.interfaces_to_user_output.keys())}. "
                f"You need to access the parameter for each output individually.\n"
            )
            if len(name) > 0:
                for k in self.interfaces_to_user_output:
                    s += f"ACTOR.{k}.{name}\n"
                s += "... where ACTOR is your actor object."
            s += f"You can still use the shortcut {name} to *set* the parameter to *all* outputs of this actor. "
            fatal(s)
        return func(self, *args)

    return _with_check


def make_property_function(interface_name):
    def p(self):
        interface_to_get = interface_name
        return self.interfaces_to_user_output[interface_to_get]

    return p


class ActorBase(GateObject):

    # hints for IDE
    attached_to: str
    filters: list
    filters_boolean_operator: str
    priority: int

    user_info_defaults = {
        "attached_to": (
            __world_name__,
            {
                "doc": "Name of the volume to which the actor is attached.",
                "setter_hook": _setter_hook_attached_to,
            },
        ),
        "mother": (
            None,
            {
                "deprecated": "The user input parameter 'mother' is deprecated. Use 'attached_to' instead. ",
            },
        ),
        "filters": (
            [],
            {
                "doc": "Filters used by this actor. ",
            },
        ),
        "filters_boolean_operator": (
            "and",
            {
                "doc": "Boolean operator to join multiple filters of this actor. ",
                "allowed_values": (
                    "and",
                    "or",
                ),
            },
        ),
        "priority": (
            100,
            {
                "doc": "Indicates where the actions of this actor should be placed "
                "in the list of all actions in the simulation. "
                "Low values mean 'early in the list', large values mean 'late in the list'. "
            },
        ),
    }

    # this dictionary is filled by the developer in each inheriting actor class
    user_output_config = {}
    # this dictionary is filled automatically during the class manufacturing process triggered by __process_this__
    # Do not redefine this manually!
    _processed_user_output_config = {}

    # private list of property names for interfaces already defined
    # in any actor (assuming all actors inherit from the base class)
    # The list is filled automatically during the class manufacturing process triggered by __process_this__
    # Do not redefine this in inheriting classes!
    _existing_properties_to_interfaces = []

    @classmethod
    def _process_user_output_config(cls):
        # it is important to create a new dictionary for this class
        # because we would otherwise write into the dictionary in the base class
        cls._processed_user_output_config = {}
        for output_name, output_config in cls.user_output_config.items():
            # if output_name not in cls._user_output_classes:
            try:
                actor_output_class = output_config["actor_output_class"]
            except KeyError:
                raise GateImplementationError(
                    f"In actor {cls.__name__}: "
                    f"No entry 'actor_output_class' specified "
                    f"in user_output_config for user_output {output_name}."
                )

            # default to "auto" if output_config has no key "interfaces"
            interfaces_of_this_output = output_config.get("interfaces", "auto")
            # if the GATE developer has not defined any interfaces, we create one automatically
            if interfaces_of_this_output == "auto":
                interface_name = output_name  # use the output name as interface name
                interfaces_of_this_output = {
                    interface_name: {
                        "interface_class": actor_output_class.get_default_interface_class(),
                        "item": 0,
                    }
                }
                # pick up parameters from the output config (where the GATE developer might have put them)
                # and add them to the interface config
                other_parameters = dict(
                    [
                        (k, v)
                        for k, v in output_config.items()
                        if k not in ("actor_output_class", "interfaces")
                    ]
                )
                interfaces_of_this_output[interface_name].update(other_parameters)

            # fill in default values where the developer has not defined any for the interface
            for i_name, config in interfaces_of_this_output.items():
                if "interface_class" not in config:
                    raise GateImplementationError(
                        f"Incorrectly configured interface {i_name} "
                        f"for actor output '{output_name}'"
                        f"in {cls}: No 'interface_class' specified. "
                    )
                interfaces_of_this_output[i_name]["suffix"] = i_name
                defaults = actor_output_class.get_user_info_default_values_interface(
                    **config
                )
                for k, v in defaults.items():
                    if k not in interfaces_of_this_output[i_name]:
                        interfaces_of_this_output[i_name][k] = v

            cls._processed_user_output_config[output_name] = {
                "actor_output_class": actor_output_class,
                "interfaces": interfaces_of_this_output,
            }

    @classmethod
    def __process_this__(cls):
        """This is a specialized version of the class method __process_this__ for actor classes."""
        # process user info defaults as in every GateObject class
        super().__process_user_info_defaults__()
        # do the actor specific class manufacturing
        cls._process_user_output_config()
        cls._create_interface_properties()
        # we generate the docstring only now when the interface properties have been created
        cls.__doc__ = cls.__get_docstring__()

    @classmethod
    def _create_interface_properties(cls):
        cls._existing_properties_to_interfaces = []
        # create properties in the actor class so the user can quickly access the interfaces
        for (
            output_name,
            user_output_config,
        ) in cls._processed_user_output_config.items():
            for interface_name, config in user_output_config["interfaces"].items():
                actor_output_class = user_output_config["actor_output_class"]

                if interface_name in cls._existing_properties_to_interfaces:
                    raise GateImplementationError(
                        f"An interface property with unique name {interface_name} "
                        f"already exists in this class. "
                    )

                # Check if this class already has this property and whether it is associated with an interface.
                if hasattr(cls, interface_name):
                    # before we raise an exception, we check if this property was created by a parent class
                    properties_in_bases = []
                    for c in cls.__bases__:
                        try:
                            properties_in_bases.extend(
                                c._existing_properties_to_interfaces
                            )
                        except AttributeError:
                            continue
                    # if the property was not created by a parent class, it must be another attribute of this class,
                    # and we should not override it. The interface name simply collides with an existing attribute
                    if interface_name not in properties_in_bases:
                        raise GateImplementationError(
                            f"Cannot create the property for interface '{interface_name}'\n"
                            f"associated with output class {actor_output_class} "
                            f"in actor {cls.__name__} \n"
                            f"because a property with that name already exists, "
                            f"but it is not associated with the interface. \n"
                            f"These are the exiting properties: {cls._existing_properties_to_interfaces}\n"
                            f"The developer needs to change the name of the interface "
                            f"(or user_output in case the interface is automatically generated). "
                        )

                doc_string = cls._get_docstring_for_interface(
                    output_name, interface_name
                )
                # we associate the property with this actor
                # Note: this does not yet create the actual interface instance
                #       which will be done only when an actor instance is initialized (__init__)
                setattr(
                    cls,
                    interface_name,
                    property(
                        fget=make_property_function(interface_name), doc=doc_string
                    ),
                )
                cls._existing_properties_to_interfaces.append(interface_name)

    @classmethod
    def __get_user_info_docstring__(cls):
        # This is a specialized version of this method for actors
        # which includes info about the output
        docstring = f"This actor has the following output:\n\n"
        for interface_name in cls._existing_properties_to_interfaces:
            docstring += f"* {interface_name}\n"
        docstring += "\n"
        docstring += super().__get_user_info_docstring__()
        return docstring

    @classmethod
    def _get_docstring_for_interface(cls, output_name, interface_name):
        interface_config = cls._processed_user_output_config[output_name]["interfaces"][
            interface_name
        ]
        actor_output_class = cls._processed_user_output_config[output_name][
            "actor_output_class"
        ]
        docstring = f"**{interface_name}**\n\n"
        docstring += "* Parameters and defaults:\n\n"
        defaults = actor_output_class.get_user_info_default_values_interface(
            **interface_config
        )
        for k, v in defaults.items():
            docstring += f"  * {k} = {v}\n"
        docstring += "\n"
        docstring += "* Methods: \n\n"
        docstring += interface_config["interface_class"].__get_docstring_methods__()
        docstring += "\n"
        docstring += "* Description of the parameters: \n\n"
        docstring += interface_config["interface_class"].__get_docstring_attributes__()
        docstring += "\n"
        return docstring

    def __init__(self, *args, **kwargs) -> None:
        GateObject.__init__(self, *args, **kwargs)
        # the actor engine is set by the actor engine during initialization
        self.actor_engine = None
        self.user_output = Box()
        self.interfaces_to_user_output = Box()
        self._init_user_output_instance()
        # the mother of the volume the actor is attached to will be automatically set
        self.mother_attached_to = None

    def __initcpp__(self):
        """Nothing to do in the base class."""

    def __getstate__(self):
        state_dict = super().__getstate__()
        state_dict["actor_engine"] = None
        return state_dict

    def __setstate__(self, state):
        super().__setstate__(state)
        for v in self.interfaces_to_user_output.values():
            v.belongs_to_actor = self
        self.__initcpp__()

    def configure_like(self, other_obj):
        super().configure_like(other_obj)
        # also pick up the configuration of the user output
        for k, v in self.user_output.items():
            v.configure_like(other_obj.user_output[k])

    def to_dictionary(self):
        d = super().to_dictionary()
        d["user_output"] = dict(
            [(k, v.to_dictionary()) for k, v in self.user_output.items()]
        )
        return d

    def from_dictionary(self, d):
        super().from_dictionary(d)
        # Create all actor output objects
        for k, v in d["user_output"].items():
            self.user_output[k].from_dictionary(v)

    def get_data(self, name=None, **kwargs):
        if name is not None:
            try:
                return self.interfaces_to_user_output[name].get_data(**kwargs)
            except KeyError:
                fatal(
                    f"No output '{name}' found in {self.type_name} actor '{self.name}'."
                )
        elif len(self.interfaces_to_user_output) == 1:
            list(self.interfaces_to_user_output.values())[0].get_data(**kwargs)
        elif len(self.interfaces_to_user_output) == 0:
            fatal(
                f"The {self.type_name} actor '{self.name}' does not handle any output."
            )
        else:
            fatal(
                f"The {self.type_name} actor '{self.name}' handles multiple outputs. "
                "There are 2 ways to fix this: \n"
                "1) Provide a keyword argument name=OUTPUT_NAME. \n"
                f"   Example: my_actor.get_data(name='{list(self.interfaces_to_user_output.keys())[0]}')\n"
                "2) Call get_data() via the output.\n"
                f"   Example: my_actor.{list(self.interfaces_to_user_output.keys())[0]}.get_data(). "
            )

    # *** shortcut properties ***
    @property
    @shortcut_for_single_output_actor
    def output_filename(self):
        return list(self.interfaces_to_user_output.values())[0].output_filename

    @output_filename.setter
    def output_filename(self, filename):
        if len(self.interfaces_to_user_output) > 1:
            for k, v in self.interfaces_to_user_output.items():
                v.output_filename = insert_suffix_before_extension(
                    filename, v.suffix, suffix_separator="_"
                )
        else:
            list(self.interfaces_to_user_output.values())[0].output_filename = filename

    @property
    @shortcut_for_single_output_actor
    def write_to_disk(self):
        return list(self.interfaces_to_user_output.values())[0].write_to_disk

    @write_to_disk.setter
    def write_to_disk(self, write_to_disk):
        for k, v in self.interfaces_to_user_output.items():
            v.write_to_disk = write_to_disk

    @property
    @shortcut_for_single_output_actor
    def keep_data_per_run(self):
        return list(self.interfaces_to_user_output.values())[0].keep_data_per_run

    @keep_data_per_run.setter
    def keep_data_per_run(self, keep_data_per_run):
        for k, v in self.interfaces_to_user_output.items():
            v.keep_data_per_run = keep_data_per_run

    def get_output_path(self, name=None, **kwargs):
        if name is not None:
            try:
                return self.interfaces_to_user_output[name].get_output_path(**kwargs)
            except KeyError:
                fatal(
                    f"No output called '{name}' found in {self.type_name} actor '{self.name}'."
                )
        elif len(self.interfaces_to_user_output) == 1:
            return list(self.interfaces_to_user_output.values())[0].get_output_path(
                **kwargs
            )
        elif len(self.interfaces_to_user_output) == 0:
            fatal(
                f"The {self.type_name} actor '{self.name}' does not handle any output."
            )
        else:
            fatal(
                f"The {self.type_name} actor '{self.name}' handles multiple outputs. "
                "There are 2 ways to fix this: \n"
                "1) Provide a keyword argument name=OUTPUT_NAME. \n"
                f"   Example: my_actor.get_output_path(name='{list(self.interfaces_to_user_output.keys())[0]}')\n"
                "2) Call get_output_path() via the output.\n"
                f"   Example: my_actor.{list(self.interfaces_to_user_output.keys())[0]}.get_output_path(). "
            )

    @property
    def actor_manager(self):
        return self.simulation.actor_manager

    @property
    def attached_to_volume(self):
        return self.simulation.volume_manager.get_volume(self.attached_to)

    def close(self):
        # first, Close the cpp part of the actor
        self.Close()
        # close all outputs
        for uo in self.user_output.values():
            uo.close()
        # remove the g4 objects and the actor engine pointer
        for v in self.__dict__:
            if "g4_" in v:
                self.__dict__[v] = None
        self.actor_engine = None
        # close the base GateObject
        super().close()

    def initialize(self):
        """This base class method initializes common settings and should be called in all inheriting classes."""

        # set the mother of the attached_to volume to the actor
        # but only if attached to a single volume.
        if isinstance(self.attached_to, str):
            vol = self.simulation.volume_manager.get_volume(self.attached_to)
            self.mother_attached_to = vol.mother
            if vol.mother is None:
                # the mother of the world is the world (sic)
                self.mother_attached_to = __world_name__
        else:
            self.mother_attached_to = "None"
        # set the name of the attached_to mother volume to cpp
        self.SetMotherAttachedToVolumeName(self.mother_attached_to)

        any_active = False
        for p in self._existing_properties_to_interfaces:
            interface = getattr(self, p)
            any_active |= interface.active
        if len(self.user_output) > 0 and not any_active:
            self.warn_user(f"The actor {self.name} has no active output. ")

        for k, v in self.user_output.items():
            v.initialize()

        # initialize filters
        try:
            self.fFilters = self.filters
        except AttributeError:
            fatal(
                f"Implementation error: Unable to set the attribute 'fFilters' in actor '{self.name}' "
                f"(actor type: {self.type_name}). "
                f"Does the actor class somehow inherit from GateVActor (as it should)?"
            )

    def _init_user_output_instance(self):
        for output_name, output_config in self._processed_user_output_config.items():
            try:
                interfaces = output_config["interfaces"]
            except AttributeError:
                raise GateImplementationError(
                    f"No interfaces found for output {output_name}. "
                )
            actor_output_class = output_config["actor_output_class"]
            # get the names of the parameters of this output class
            # we do not need to specify the item (in case the output handles a container)
            # because the names are equal for all data items in the container
            default_params = list(
                actor_output_class.get_user_info_default_values_interface().keys()
            )

            # create and add the instance of the actor output
            self._add_user_output(actor_output_class, output_name)

            # now create the interface instances linking to the actor output
            for interface_name, interface_config in interfaces.items():
                interface_params = dict(
                    [(_k, _v) for _k, _v in interface_config.items()]
                )
                try:
                    interface_class = interface_params.pop("interface_class")
                except KeyError:
                    raise GateImplementationError(
                        f"Incorrectly configured interface {interface_name} "
                        f"for actor output {output_name}"
                        f"in {self.type_name}: No 'interface_class' specified. "
                    )
                self._add_interface_to_user_output(
                    interface_class, output_name, interface_name, **interface_params
                )
                # use the newly created interface to set the defaults
                interface = self.interfaces_to_user_output[interface_name]
                for p in default_params:
                    v = interface_config[p]
                    setattr(interface, p, v)

    def _add_user_output(
        self,
        actor_output_class,
        name,
        **kwargs,
    ):
        """Method to be called internally (not by user) in the specific actor class implementations."""

        if (actor_output_class.__name__ == ActorOutputRoot.__name__) and any(
            [
                type(v).__name__ == ActorOutputRoot.__name__
                for v in self.user_output.values()
            ]
        ):
            raise GateImplementationError("Only one ROOT output per actor supported. ")

        self.user_output[name] = actor_output_class(
            name=name,
            simulation=self.simulation,
            belongs_to=self,
            **kwargs,
        )

    def _add_interface_to_user_output(
        self, interface_class, user_output_name, interface_name, **kwargs
    ):
        if interface_name not in self.interfaces_to_user_output:
            self.interfaces_to_user_output[interface_name] = interface_class(
                self, user_output_name, interface_name, **kwargs
            )
            # set an instance attribute __doc__ for this interface
            # equivalent to the property linking to this interface
            # so that the user gets a meaningful info when using something like
            # ``help(my_dose_actor.dose_uncertainty)
            self.interfaces_to_user_output[interface_name].__doc__ = getattr(
                type(self), interface_name
            ).__doc__
        else:
            raise GateImplementationError(
                f"An actor output user interface called '{interface_name}' already exists. "
            )

    def recover_user_output(self, actor):
        self.user_output = actor.user_output
        # if actor comes from a subprocess, its reference actor.simulation
        # points to the simulation instance created in the subprocess
        # Therefore, we have to reset it to the right simulation instance,
        # namely the one from the main process
        for u in self.user_output.values():
            u.simulation = self.simulation
        for v in self.interfaces_to_user_output.values():
            v.belongs_to_actor = self

    def store_output_data(self, output_name, run_index, *data):
        self._assert_output_exists(output_name)
        self.user_output[output_name].store_data(run_index, *data)

    def write_output_to_disk_if_requested(self, output_name):
        self._assert_output_exists(output_name)
        self.user_output[output_name].write_data_if_requested()

    def _assert_output_exists(self, output_name):
        if output_name not in self.user_output:
            fatal(f"No output named '{output_name}' found for actor {self.name}.")

    def get_output_path_for_item(self, output_name, which, item):
        return self.user_output[output_name].get_output_path(which, item)

    def get_output_path_string(self, **kwargs):
        return str(self.get_output_path(**kwargs))

    def get_output_path_for_item_string(self, output_name, which, item):
        return str(self.user_output[output_name].get_output_path(which, item))

    def StartSimulationAction(self):
        """Default virtual method for inheritance"""
        pass

    def EndSimulationAction(self):
        """Default virtual method for inheritance"""
        pass


process_cls(ActorBase)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
from test095_spect_helpers import *
import json


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="", output_folder="test095"
    )

    p = nm670.get_geometrical_parameters()
    collimators = p.collimators
    # collimators = ["lehr"]
    n = len(collimators)
    collimators.append(collimators[0])
    is_ok = True

    crystal_sizes = ["3/8", "5/8"]

    for crystal_size in crystal_sizes:
        i = 0
        for colli_type in collimators:
            # rotate the gantry only for the last collimator
            if i == n:
                radius = 20 * gate.g4_units.cm
                rotation = 12
            else:
                radius = 0
                rotation = 0

            sim, fn = build_test_simu(
                paths, "nm670", colli_type, rotation, radius, crystal_size
            )

            # go
            sim.run(start_new_process=True)

            # compare the json files
            ref_json_file = paths.output_ref / fn
            j_ref = json.load(open(ref_json_file))
            j_test = json.load(open(sim.user_hook_after_init_arg))
            added, removed, modified, same = utility.dict_compare(
                j_ref,
                j_test,
                tolerance=1e5,
                ignored_keys=[
                    "copy_no",
                ],
            )
            b = len(added) == 0 and len(removed) == 0 and len(modified) == 0
            utility.print_test(b, f"Compare json volume info with reference")
            is_ok = is_ok and b
            i = i + 1

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.sources.phidsources as phidsources
import click
import numpy as np
import matplotlib.pyplot as plt
import radioactivedecay as rd
from opengate.utility import g4_best_unit

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("rad_name", nargs=1)
@click.option("--output", "-o", default=None, help="output file")
@click.option("--verbose", "-v", is_flag=True, default=False, help="verbose")
@click.option(
    "--log_scale", "--log", is_flag=True, default=False, help="plot with log scale"
)
@click.option("--start_h", default=24.0, help="time start in hour")
@click.option("--duration_s", default=200.0, help="duration in sec")
@click.option("--n_first", "-n", default=10, help="print n largest weights")
@click.option(
    "--half_life", default=None, type=float, help="Specify the half-life for metastable"
)
@click.option(
    "--half_life_unit",
    default=None,
    help="Specify the half-life unit for metastable us, s, h etc",
)
def go(
    rad_name,
    output,
    verbose,
    log_scale,
    start_h,
    duration_s,
    n_first,
    half_life,
    half_life_unit,
):
    """
    Display all gammas (isomeric transition + atomic relaxation) for a radionuclide with all its decay chains.
    The relative activities of the daughters radionuclides are computed according to start+duration timing.

    For metastable, you can set the half-life.

    """

    # get nuclide, a, z
    nuclide = phidsources.get_nuclide_from_name(rad_name)
    name = nuclide.nuclide[: nuclide.nuclide.index("-")]

    # get all daughters
    daughters = phidsources.get_nuclide_progeny(nuclide)
    if verbose:
        print(
            f"Found {len(daughters)} daughters for {name} HL={nuclide.half_life('m')} min"
        )

    # TAC
    inv = rd.Inventory({nuclide.nuclide: 1.0}, "Bq")
    inv = inv.decay(start_h, "h")
    if verbose:
        for a in inv.activities():
            print(f"{a} {inv.activities()[a] * 100:.2f}%")

    ca = inv.cumulative_decays(duration_s, "s")
    for n in ca:
        ca[n] /= duration_s
    if verbose:
        for a in inv.activities():
            print(f"{a} {inv.activities()[a] * 100:.2f}%")

    # Isomeric transition
    sec = gate.g4_units.s
    us = gate.g4_units.us
    hour = gate.g4_units.h
    if half_life is None:
        it = phidsources.isomeric_transition_load_all_gammas(nuclide)
        half_life = nuclide.half_life("s") * sec
    else:
        unit = gate.g4_units[half_life_unit]
        hl = half_life * unit
        print(f"Half life is {g4_best_unit(hl, 'Time')}")
        it = phidsources.isomeric_transition_load_all_gammas(nuclide, half_life=hl)

    # Atomic relaxation
    ar = phidsources.atomic_relaxation_load_all_gammas(nuclide)

    # merge both
    all = it + ar

    # take tac into account
    for x in all:
        n = x["nuclide"].nuclide.nuclide
        if n in ca:
            p = ca[n]
            x["intensity"] *= p
        else:
            x["intensity"] = 0

    # prepare to sort
    w_all = np.array([w["intensity"] for w in all])
    ene_all = np.array([w["energy"] for w in all])
    t_all = np.array([w["type"] for w in all])
    n_all = np.array([w["nuclide"] for w in all])

    # sort by weights
    sorted_indices = np.argsort(-w_all)
    sorted_weights = w_all[sorted_indices]
    sorted_ene = ene_all[sorted_indices]
    sorted_types = t_all[sorted_indices]
    sorted_nuc = n_all[sorted_indices]

    # print
    keV = gate.g4_units.keV
    i = 0
    n = n_first
    if verbose:
        print(f"Number of IT = {len(it)} and of AR = {len(ar)}")
        for e, w, t, nuc in zip(
            sorted_ene[:n], sorted_weights[:n], sorted_types[:n], sorted_nuc[:n]
        ):
            print(
                f"{i} Energy {e / keV:.3f} keV     {w * 100:.3f} %  {t} {nuc.nuclide.nuclide}"
            )
            i += 1

    # plot
    SIZE = 14
    fig, ax = plt.subplots(1, 1, figsize=(15, 5))
    color_mapping = {"it": "red", "ar": "blue"}
    colors = [color_mapping[val] for val in sorted_types]
    w = (sorted_ene.max() - sorted_ene.min()) / 300
    print(f"Max energy = {sorted_ene.max() / keV} keV")
    ax.bar(
        sorted_ene / keV,
        sorted_weights * 100,
        width=w / keV,
        color=colors,
        log=log_scale,
    )
    l = len(sorted_ene)

    # Add text annotations for the three largest bars
    for i in range(0, n_first):
        energy = sorted_ene[i] / keV
        w = sorted_weights[i] * 100
        n = sorted_nuc[i].nuclide.nuclide
        text = f"{energy:.1f} keV\n{w:.1f}% \n {n}"
        plt.annotate(
            text,
            xy=(energy, w),
            xytext=(energy, w),
            ha="center",
            va="bottom",
            fontsize=10,
            bbox=dict(boxstyle="round,pad=0.2", fc="white", ec="none"),
        )

    ax.set_xlabel("Energy keV", fontsize=SIZE)
    ax.set_ylabel("Intensity %", fontsize=SIZE)
    plt.xticks(fontsize=SIZE)
    plt.yticks(fontsize=SIZE)

    x_range = ax.get_xlim()
    y_range = ax.get_ylim()
    x = (x_range[1] - x_range[0]) / 5 * 3.5 + x_range[0]
    y = (y_range[1] - y_range[0]) / 2 + y_range[0]
    plt.annotate(
        f"{nuclide.nuclide}\n"
        f"Half-life = {g4_best_unit(half_life, 'Time')}\n"
        f"at t={g4_best_unit(start_h * hour, 'Time')}\n"
        f"duration t={g4_best_unit(duration_s*sec, 'Time')}\n"
        f"{l} photon sources \n",
        xy=(x, y),
        fontsize=SIZE,
    )
    # ax.legend()

    if output:
        print(f"Output plot in {output}")
        fig.savefig(output)
    else:
        plt.show()


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.sources.utility import get_spectrum
import numpy as np
import gatetools

import matplotlib.pyplot as plt


def run_simulation(paths):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    sim.random_seed = 987654321
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    g_cm3 = gate.g4_units.g_cm3

    # materials
    sim.volume_manager.material_database.add_material_weights(
        "Vacuum", ["H"], [1], 1e-12 * g_cm3
    )

    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "Vacuum"

    phsp = sim.add_volume("Sphere", "phsp")
    phsp.rmin = 0.1 * m
    phsp.rmax = 1 * m
    phsp.material = world.material

    source = add_source_energy_spectrum_discrete(sim, phsp)

    # actors
    phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp_actor")
    phsp_actor.output_filename = "test010_energy_spectrum_discrete.root"
    phsp_actor.attached_to = phsp.name
    phsp_actor.attributes = [
        "KineticEnergy",
    ]

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # get info
    ekin = root_load_ekin(str(phsp_actor.get_output_path()))

    # test
    energies = source.energy.spectrum_energies
    weights = source.energy.spectrum_weights

    src_data = np.array(weights)
    src_data = src_data / np.sum(src_data)

    energy_counts = {energy: 0 for energy in energies}
    for energy in ekin:
        if energy not in energy_counts:
            energy_counts[energy] = 0
        energy_counts[energy] += 1

    sim_data = [energy_counts[x] for x in energies]
    sim_data = sim_data / np.sum(sim_data)

    relerrs = (sim_data - src_data) / src_data
    oks = [
        utility.check_diff_abs(0, abs(relerr), tolerance=0.05, txt="relative error")
        for relerr in relerrs
    ]
    is_ok = all(oks)

    plot(
        paths.output / "test010_plot_energy_spectrum_discrete.png",
        energies,
        src_data,
        sim_data,
        relerrs,
    )

    utility.test_ok(is_ok)


def plot(output_file, energies, src_data, sim_data, relerrs):
    fig, ax = plt.subplots(figsize=(8.5, 6))
    ax.set_xlabel("Energy (MeV)")
    ax.set_ylabel("Number of particles")
    ax.scatter(energies, src_data, label="input energy spectrum", marker="o")
    ax.scatter(
        energies, sim_data, label="simulated energy spectrum", marker=".", linewidth=0.9
    )

    ax2 = ax.twinx()
    ax2.set_ylabel("Relative error")
    ax2.set_ylim([-0.05, +0.05])
    ax2.plot(energies, relerrs, label="relative error", linewidth=0.4, color="C2")

    ax.legend()

    fig.savefig(output_file, dpi=300)
    plt.close(fig)


def root_load_ekin(root_file: str):
    data_ref, keys_ref, _ = gatetools.phsp.load(root_file)

    index_ekin = keys_ref.index("KineticEnergy")
    ekin = [data_ref_i[index_ekin] for data_ref_i in data_ref]

    return ekin


def add_source_energy_spectrum_discrete(sim, phsp):
    spectrum = get_spectrum("Lu177", "gamma")

    source = sim.add_source("GenericSource", "beam")
    source.attached_to = phsp.name
    source.particle = "gamma"
    source.n = 5e5 / sim.number_of_threads
    source.position.type = "point"
    source.direction.type = "iso"
    source.energy.type = "spectrum_discrete"
    source.energy.spectrum_energies = spectrum.energies
    source.energy.spectrum_weights = spectrum.weights

    return source


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "test010_generic_source_energy_spectrum", output_folder="test010"
    )

    run_simulation(paths)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.tests.utility as tu

if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, "", "test074")

    # create simulation
    sim = gate.Simulation()
    sim.g4_verbose = True
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    eV = gate.g4_units.eV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # set the world size like in the Gate macro
    sim.world.size = [3 * m, 3 * m, 3 * m]

    # add a simple crystal volume
    crystal = sim.add_volume("Box", "crystal")
    crystal.size = [3 * mm, 3 * mm, 20 * mm]
    crystal.translation = [0 * cm, 0 * cm, 0 * cm]
    crystal.material = "BGO"

    # add a surface
    # Users can specify their own path surface properties file by
    # sim.physics_manager.surface_properties_file = (
    #     "/Users/data_machine/Work/pyGate/opengate/opengate/data/SurfaceProperties.xml"
    # )
    # By default, Gate uses the file opengate/data/SurfaceProperties.xml

    # Syntax to add optical surface between two volumes
    # add_optical_surface("volume_from","volume_to", "surface_name")
    # keywords can be used for clarity
    # the name of the GATE optical surface is created automatically
    opt_surf_world_crystal = sim.physics_manager.add_optical_surface(
        volume_from="world",
        volume_to="crystal",
        g4_surface_name="polished_teflon_wrapped",
    )
    opt_surf_crystal_world = sim.physics_manager.add_optical_surface(
        "crystal", "world", "Rough_LUT"
    )

    print(sim.physics_manager.dump_optical_surfaces())

    # Examples -
    # sim.add_optical_surface("OpticalSystem", "Crystal1","PolishedTeflon_LUT")
    # sim.add_optical_surface("Crystal1", "OpticalSystem", "PolishedTeflon_LUT")
    # sim.add_optical_surface("Greasepixel", "Crystal1", "Polished_LUT")
    # sim.add_optical_surface("Crystal1", "Greasepixel", "Polished_LUT")
    # sim.add_optical_surface("Greasepixel", "pixel", "Detector_LUT")
    # sim.add_optical_surface("pixel", "Greasepixel", "Detector_LUT")

    # change physics
    # For the generation of Cerenkov, physics_list_name must
    # be set to G4EmStandardPhysics_option4 and production cuts
    # of electron must be set to 0.1 mm (Reason unknown)
    # Reference - https://opengate.readthedocs.io/en/latest/generating_and_tracking_optical_photons.html
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.set_production_cut("crystal", "electron", 0.1 * mm)
    sim.physics_manager.energy_range_min = 10 * eV
    sim.physics_manager.energy_range_max = 1 * MeV
    sim.physics_manager.special_physics_constructors.G4OpticalPhysics = True

    # Users can specify their own path optical properties file by
    # sim.physics_manager.optical_properties_file = PATH_TO_FILE
    # By default, Gate uses the file opengate/data/OpticalProperties.xml

    # Change source
    source = sim.add_source("GenericSource", "gamma1")
    source.particle = "gamma"
    source.energy.mono = 0.511 * MeV
    source.activity = 10 * Bq
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.position.translation = [0 * cm, 0 * cm, 2.2 * cm]

    # add phase actor
    phase = sim.add_actor("PhaseSpaceActor", "Phase")
    phase.attached_to = crystal
    phase.attributes = [
        "Position",
        "PostPosition",
        "PrePosition",
        "ParticleName",
        "TrackCreatorProcess",
        "EventKineticEnergy",
        "KineticEnergy",
        "PDGCode",
    ]
    phase.output_filename = "test070_pet_surfaces_1.root"

    sim.user_hook_after_run = gate.userhooks.user_hook_dump_material_properties
    sim.run()

    is_ok = all(t is True for t in sim.user_hook_log)
    tu.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test053_phid_helpers2 import *
import os
import opengate as gate

if __name__ == "__main__":
    paths = get_default_test_paths(__file__, "", output_folder="test053")

    # bi213 83 213
    # ac225 89 225
    # fr221 87 221
    # pb 82 212
    # po 84 213
    # tl 81 209
    z = 89
    a = 225
    nuclide, _ = get_nuclide_and_direct_progeny(z, a)
    print(nuclide)

    sim = gate.Simulation()
    sim_name = f"{nuclide.nuclide}_10_ref"
    create_sim_test053(sim, sim_name, output=paths.output)

    phsp = sim.get_actor("phsp")
    phsp.filters = [phsp.filters[0]]
    print(phsp.output_filename)

    mm = g4_units.mm
    sim.physics_manager.global_production_cuts.all = 1 * mm

    # sources
    activity_in_Bq = 500
    add_source_generic(sim, z, a, activity_in_Bq)

    # timing
    sec = g4_units.second
    min = g4_units.minute
    start_time = 15 * min
    end_time = start_time + 2 * min
    duration = end_time - start_time
    print(f"start time {start_time / sec}")
    print(f"end time {end_time / sec}")
    print(f"Duration {duration / sec}")
    print(f"Ions {activity_in_Bq * duration / sec:.0f}")
    sim.run_timing_intervals = [[0, end_time]]

    # go
    sim.run(start_new_process=True)

    # print stats
    stats = sim.get_actor("stats")
    print(stats)

    # compare with reference root file
    warning(f"check root files")
    root_model = sim.get_actor("phsp").get_output_path()
    root_ref = paths.output_ref / os.path.basename(root_model)
    keys = ["KineticEnergy", "TrackCreatorModelIndex"]
    tols = [0.005, 3.4]
    img = paths.output / str(root_model).replace(".root", ".png")
    is_ok = compare_root3(
        root_ref, root_model, "phsp", "phsp", keys, keys, tols, None, None, img
    )

    test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
from opengate.geometry.volume_info import *
import json

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("json_file", type=click.Path(exists=True))
@click.option("--axis", "-a", default="all", help="all, XY or XZ or YZ")
@click.option("--hl", default=None, multiple=True, help="highlight volumes")
def go(json_file, axis, hl):
    # Load the JSON data
    with open(json_file, "r") as f:
        spect_data = json.load(f)

    # Plot in XY plane (2D view)
    if axis == "all":
        plot_all_views_2d(spect_data, show_labels=True, highlight_volumes=hl)
    else:
        plot_volume_boundaries_2d(
            spect_data, plane=axis, show_labels=True, highlight_volumes=hl
        )


if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

from scipy.spatial.transform import Rotation


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test008_dose_actor", output_folder="test012"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 2
    sim.random_seed = 123456789
    sim.check_volumes_overlap = True
    sim.output_dir = paths.output

    # shortcuts to units
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    m = gate.g4_units.m
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.rotation = Rotation.from_euler("x", 10, degrees=True).as_matrix()
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", 20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 150 * MeV
    source.particle = "proton"
    source.position.radius = 1 * nm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    # source.activity = 2e5 / sim.number_of_threads * Bq
    source.activity = 5e4 / sim.number_of_threads * Bq

    """
    It needs at least around 2e5 particles for the multithread to be faster than mono thread
    """

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.edep.output_filename = "test012-edep.mhd"
    dose_actor.attached_to = "waterbox"
    dose_actor.size = [99, 99, 99]
    dose_actor.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose_actor.translation = [2 * mm, 3 * mm, -2 * mm]

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True

    # print info
    sim.volume_manager.print_volumes()

    # verbose
    # sim.g4_commands_after_init.append('/tracking/verbose 0')
    sim.g4_commands_after_init.append("/run/verbose 2")
    # sim.g4_commands_after_init.append("/event/verbose 2")
    # sim.g4_commands_after_init.append("/tracking/verbose 1")

    # start simulation
    sim.run()

    # print results at the end
    print(stat)
    print(dose_actor)

    # tests
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    # change the number of run to the number of threads
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stat, stats_ref, 0.10)

    is_ok = (
        utility.assert_images(
            paths.gate_output / "output-Edep.mhd",
            dose_actor.edep.get_output_path(),
            stat,
            tolerance=45,
        )
        and is_ok
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.contrib.spect.pytomography_helpers import *
from opengate.tests import utility

if __name__ == "__main__":

    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test094_pytomography"
    )
    data_folder = paths.data

    # units
    mm = g4_units.mm
    cm = g4_units.cm
    sec = g4_units.s
    keV = g4_units.keV

    # Init SPECT simulation, no phantom, no source
    sc = SPECTConfig()
    sc.output_folder = paths.output_ref
    sc.detector_config.model = "intevo"
    sc.detector_config.collimator = "melp"
    sc.detector_config.number_of_heads = 2
    sc.detector_config.size = [128, 128]
    sc.detector_config.spacing = [4.8 * mm, 4.8 * mm]
    sc.detector_config.digitizer_function = intevo.add_digitizer
    sc.detector_config.digitizer_channels = intevo.get_default_energy_windows("lu177")
    sc.phantom_config.image = data_folder / "ct_5mm.mhd"
    sc.phantom_config.translation = [0, 0, -20 * cm]
    sc.source_config.radionuclide = "177lu"
    sc.acquisition_config.radius = 420 * mm
    sc.acquisition_config.duration = 30 * sec
    sc.acquisition_config.number_of_angles = 30

    # Create the simulation to retrieve some spect information
    # (no need to run the simulation)
    sim = gate.Simulation()
    sc.setup_simulation(sim, visu=False)
    metadata = pytomography_build_metadata_and_attenuation_map(
        sc, sim, 208 * keV, output_folder=paths.output, verbose=True
    )

    # write metadata as a JSON file
    fn = paths.output / "pytomography_gate.json"
    with open(fn, "w") as f:
        json.dump(metadata, f, indent=4)
    print(f"Metadata written in {fn}")

    # check the generated file according to the reference one
    ref_json = paths.output_ref / "pytomography_gate.json"
    # compare the dict
    with open(ref_json, "r") as f:
        ref_metadata = json.load(f)

    # compare the dict
    added, removed, modified, same = utility.dict_compare(metadata, ref_metadata)
    is_ok = len(added) == 0 and len(removed) == 0 and len(modified) == 0
    utility.print_test(is_ok, f"Labels comparisons, added:    {added}")
    utility.print_test(is_ok, f"Labels comparisons, removed:  {removed}")
    utility.print_test(is_ok, f"Labels comparisons: modified: {modified}")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from pathlib import Path


def create_photon_attenuation_image(
    image_filename,
    labels_filename,
    energy,
    material_database=None,
    database="NIST",
    verbose=False,
    density_tol=None,
    progress_bar=False,
):
    # create a temporary simulation
    sim = gate.Simulation()
    sim.verbose_level = gate.logger.NONE
    sim.progress_bar = progress_bar

    # add the voxelized image
    image_volume = sim.add_volume("Image", "image")
    image_volume.image = image_filename
    image_volume.material = "G4_AIR"

    # labels
    if labels_filename is None:
        if density_tol is None:
            density_tol = 0.05 * gate.g4_units.g_cm3
        f1 = gate.utility.get_data_folder() / "Schneider2000MaterialsTable.txt"
        f2 = gate.utility.get_data_folder() / "Schneider2000DensitiesTable.txt"
        vm, materials = gate.geometry.materials.HounsfieldUnit_to_material(
            sim, density_tol, f1, f2
        )
        image_volume.voxel_materials = vm
        filename = Path(image_filename)
        labels_filename = filename.parent / f"{filename.stem}_labels.json"
        material_database = filename.parent / f"{filename.stem}_materials.db"
        image_volume.write_material_database(material_database)
        image_volume.write_label_to_material(labels_filename)

    image_volume.read_label_to_material(labels_filename)

    # material
    if material_database is not None:
        sim.volume_manager.add_material_database(material_database)

    # mu map actor (process at the first begin of run only)
    mumap = sim.add_actor("AttenuationImageActor", "mumap")
    mumap.image_volume = image_volume
    mumap.energy = energy
    mumap.database = database
    mumap.write_to_disk = False
    verbose and print(f"Energy is {mumap.energy / gate.g4_units.keV} keV")
    verbose and print(f"Database is {mumap.database}")

    # go
    verbose and print("Starting computing mu ...")
    sim.run(start_new_process=True)

    # retrieve the created image
    im = mumap.attenuation_image.merged_data.image
    return im



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.sources.phidsources import *
import click
import matplotlib.pyplot as plt

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("rad_name", nargs=1)
@click.option("--output", "-o", default=None, help="output file")
@click.option(
    "--verbose/--no-verbose", "-v", is_flag=True, default=False, help="Print data"
)
@click.option("--plot/--no-plot", is_flag=True, default=True, help="Plot data")
@click.option(
    "--log_scale", "--log", is_flag=True, default=False, help="plot with log scale"
)
def go(rad_name, output, verbose, plot, log_scale):
    """
    Display the isomeric gamma spectra for one given radionuclide.

    Data are retrieved from Geant4 database and can be checked with
    http://www.lnhb.fr/nuclear-data/module-lara/
    https://www-nds.iaea.org/relnsd/vcharthtml/VChartHTML.html
    By default, data are stored in text files in the folder: opengate/opengate/data/isomeric_transition
    """
    # FIXME dont work with some metasable state (Tc99m)

    # get nuclide
    nuclide = get_nuclide_from_name(rad_name)

    # load the isomeric transition data
    ene, weights = isomeric_transition_load(nuclide)

    # print
    keV = g4_units.keV
    if verbose:
        i = 0
        for e, w in zip(ene, weights):
            print(f"{i} Energy {e / keV:.3f} keV     {w * 100:.3f} %")
            i += 1
        print()

    # plot
    if plot:
        f, ax = plt.subplots(1, 1, figsize=(15, 5))
        ax.bar(
            ene / keV,
            height=weights * 100,
            width=0.4,
            label=f"{nuclide.nuclide}",
            log=log_scale,
        )
        ax.set_xlabel("Energy in keV")
        ax.set_ylabel("Intensity in %")
        ax.legend()
        if output:
            print(f"Output plot in {output}")
            f.savefig(output)
        else:
            plt.show()


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.userhooks import check_production_cuts
from test013_phys_lists_helpers import create_pl_sim


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test013_phys_lists")

    mm = gate.g4_units.mm

    # create simulation
    sim = create_pl_sim()

    # keep only ion sources
    sim.source_manager.sources.pop("gamma")

    # change physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.enable_decay = True

    sim.physics_manager.global_production_cuts.gamma = 5 * mm
    sim.physics_manager.global_production_cuts.electron = "default"
    sim.physics_manager.global_production_cuts.positron = 3 * mm
    sim.physics_manager.global_production_cuts.proton = 1 * mm

    sim.physics_manager.set_production_cut(
        volume_name="waterbox",
        particle_name="gamma",
        value=2 * mm,
    )
    sim.physics_manager.set_production_cut(
        volume_name="b2",
        particle_name="electron",
        value=5 * mm,
    )

    # em parameters
    sim.physics_manager.em_parameters.fluo = True
    sim.physics_manager.em_parameters.auger = True
    sim.physics_manager.em_parameters.auger_cascade = True
    sim.physics_manager.em_parameters.pixe = True
    sim.physics_manager.em_parameters.deexcitation_ignore_cut = True

    sim.physics_manager.em_switches_world.deex = True
    sim.physics_manager.em_switches_world.auger = True
    sim.physics_manager.em_switches_world.pixe = True

    print("Phys list cuts:")
    print(sim.physics_manager.dump_production_cuts())

    # start simulation
    # sim.g4_commands_after_init.append("/tracking/verbose 1")
    sim.g4_verbose = False
    sim.user_hook_after_init = check_production_cuts
    sim.run()

    stats = sim.get_actor("Stats")

    # gate_test4_simulation_stats_actor
    # Gate mac/main_4.mac
    f = paths.gate_output / "stat_5.txt"
    print("Reference file", f)
    stats_ref = utility.read_stat_file(f)
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.13)

    utility.test_ok(is_ok)



from opengate.utility import g4_units
import numpy as np
import pydicom


def liste_CP(file, beam_sequence_ID):
    ds = pydicom.dcmread(file)[0x300A, 0x00B0].value[beam_sequence_ID][0x300A, 0x0111]
    return len(ds.value)


def list_of_beam_sequence_ID(file):
    patient_info = pydicom.dcmread(file)[0x300A, 0x0070]
    nb_of_beam_seq = len(patient_info.value[0][("0x300C", "0x0004")].value)
    id_of_beam_sequence = []
    for i in range(nb_of_beam_seq):
        number_of_mu = retrieve_value_in_DICOM_RTPlan(
            patient_info.value[0],
            [[("0x300C", "0x0004"), i], [("0x300A", "0x0086"), None]],
        )
        if number_of_mu is not None:
            id_of_beam_sequence.append(i)
    return id_of_beam_sequence


def read(file, cp_id="all_cp", arc_id=None):
    mm = g4_units.mm
    jaws_1 = 0
    jaws_2 = 1
    angle = 2
    dir_angle = 3
    isocenter = 4
    dose_weights = 5
    collimation_angle = 6
    leaves = 7
    Mu_number = 8

    jaw_1_array = []
    jaw_2_array = []
    rot_angle_array = []
    isocenter_array = []
    dose_weight_array = []
    collimation_angle_array = []
    leaf_array = []
    if arc_id is None:
        id_of_beam_sequence = list_of_beam_sequence_ID(file)
    else:
        arc_id = int(arc_id)
        id_of_beam_sequence = [arc_id]

    for beam_seq_id in id_of_beam_sequence:
        l_jaws_1 = []
        l_jaws_2 = []
        l_angle = []
        l_dir_angle = []
        l_isocenter = []
        x_leaf = []
        l_dose_weights = []
        l_collimation_angle = []
        data_set = extract_dataset(file, beam_seq_id)
        number_of_MU = data_set[Mu_number]["MU number"]
        key_list = [
            "jaws_1",
            "jaws_2",
            "rot_angle",
            "dir_angle",
            "isocenter",
            "cumulative weight",
            "collimation angle",
        ]
        l_parameters = [
            l_jaws_1,
            l_jaws_2,
            l_angle,
            l_dir_angle,
            l_isocenter,
            l_dose_weights,
            l_collimation_angle,
        ]
        if cp_id == "all_cp":
            nb_cp_id = len(data_set[jaws_1]["jaws_1"])
            l_cp_id = np.arange(0, nb_cp_id, 1)
        for id in l_cp_id:
            for i, key in enumerate(key_list):
                if data_set[i][key][id] is not None:
                    if key == "isocenter":
                        l_parameters[i].append(np.array(data_set[i][key][id]) * mm)
                    elif (
                        key != "rot_angle"
                        and key != "dir_angle"
                        and key != "collimation angle"
                    ):
                        l_parameters[i].append(float(data_set[i][key][id]) * mm)
                    else:
                        l_parameters[i].append(data_set[i][key][id])
                else:
                    if id != 0:
                        if data_set[i][key][id] is None:
                            if l_parameters[i][id - 1] is not None:
                                l_parameters[i].append(l_parameters[i][id - 1])
                        else:
                            l_parameters[i].append(data_set[i][key][id])
                    else:
                        l_parameters[i].append(data_set[i][key][id])

        tmp_dir_angle = []
        for i in range(len(l_dir_angle)):

            if l_dir_angle[i] == "CW":
                tmp_dir_angle.append(1)
            elif l_dir_angle[i] == "CC":
                tmp_dir_angle.append(-1)
            else:
                tmp_dir_angle.append(1)

        l_dir_angle = np.array(tmp_dir_angle)

        for id in l_cp_id:
            leaf_block_1 = []
            leaf_block_2 = []
            leaf_blocks = [leaf_block_1, leaf_block_2]
            for i in range(len(data_set[leaves][0])):
                for side_ID in range(2):
                    if data_set[leaves][0][i]["leaves"][id] is not None:
                        leaf_blocks[side_ID].append(
                            data_set[leaves][side_ID][i]["leaves"][id]
                        )
                    else:
                        if id != 0:
                            if data_set[leaves][0][i]["leaves"][id - 1] is not None:
                                leaf_blocks[side_ID].append(
                                    data_set[leaves][side_ID][i]["leaves"][id - 1]
                                )
                            else:
                                leaf_blocks[side_ID].append(None)
                        else:
                            leaf_blocks[side_ID].append(None)
            if None not in leaf_block_1 and None not in leaf_block_2:
                leaf_block_tmp = np.array(leaf_block_1 + leaf_block_2, dtype=float) * mm
                x_leaf.append(leaf_block_tmp)
            else:
                x_leaf.append([None])

        if None in l_dose_weights:
            diff_dose_weights = None
        else:
            l_dose_weights = np.array(l_dose_weights)
            diff_dose_weights = np.diff(l_dose_weights) * number_of_MU
            dose_weight_array += diff_dose_weights.tolist()
            # diff_dose_weights = diff_dose_weights / np.median(diff_dose_weights)
        if None in l_angle:
            l_angle = None
        else:
            l_angle = np.array(l_angle) * l_dir_angle
            l_angle[l_angle < 0] = -l_angle[l_angle < 0]
            diff_angle = np.diff(l_angle)
            diff_angle[diff_angle > 300] = diff_angle[diff_angle > 300] - 360
            diff_angle[diff_angle < -300] = diff_angle[diff_angle < -300] + 360
            l_angle = l_angle[0:-1] + diff_angle / 2
            l_angle[l_angle > 360] -= 360
            l_angle[l_angle < 0] += 360
            rot_angle_array += l_angle.tolist()
        if None in x_leaf[0]:
            x_leaf = None
        else:
            x_leaf = np.array(x_leaf)
            x_leaf = x_leaf[0:-1] + np.diff(x_leaf, axis=0) / 2
            leaf_array += x_leaf.tolist()

        if None in l_jaws_1:
            l_jaws_1 = None
        else:
            l_jaws_1 = np.array(l_jaws_1)
            l_jaws_1 = l_jaws_1[0:-1] + np.diff(l_jaws_1) / 2
            jaw_1_array += l_jaws_1.tolist()

        if None in l_jaws_2:
            l_jaws_2 = None
        else:
            l_jaws_2 = np.array(l_jaws_2)
            l_jaws_2 = l_jaws_2[0:-1] + np.diff(l_jaws_2) / 2
            jaw_2_array += l_jaws_2.tolist()

        if None in l_collimation_angle:
            l_collimation_angle = None
        else:
            l_collimation_angle = np.array(l_collimation_angle, dtype=float)
            l_collimation_angle = (
                l_collimation_angle[0:-1] + np.diff(l_collimation_angle) / 2
            )
            collimation_angle_array += l_collimation_angle.tolist()

        isocenter_array += l_isocenter

    jaw_1_array = np.array(jaw_1_array)
    jaw_2_array = np.array(jaw_2_array)
    leaf_array = np.array(leaf_array)
    rot_angle_array = np.array(rot_angle_array)
    dose_weight_array = np.array(dose_weight_array)
    dose_weight_array = dose_weight_array / np.median(dose_weight_array)
    isocenter_array = np.array(isocenter_array)
    collimation_angle_array = np.array(collimation_angle_array)

    rt_plan_parameters = {
        "jaws 1": jaw_1_array,
        "weight": dose_weight_array,
        "jaws 2": jaw_2_array,
        "leaves": leaf_array,
        "gantry angle": rot_angle_array,
        "isocenter": isocenter_array,
        "collimation angle": collimation_angle_array,
    }
    return rt_plan_parameters


file = "/home/mjacquet/Documents/Simulation_RT_plan/patient_data/IGR/AGORL_CLB_P1toP20/AGORL_P17/RP.1.2.752.243.1.1.20200423201445736.8300.22461.dcm"


def retrieve_value_in_DICOM_RTPlan(cp, list):
    if list[0][0] in cp:
        if list[0][1] is not None:
            value = cp[list[0][0]].value[list[0][1]]
        else:
            value = cp[list[0][0]].value
        if len(list) > 1:
            return retrieve_value_in_DICOM_RTPlan(value, list[1:])
        if len(list) == 1:
            if value == "NONE":
                return None
            return value
    return None


def extract_dataset(file, beam_sequence_ID):
    jaws_1 = 0
    jaws_2 = 1
    angle = 2
    dir_angle = 3
    isocenter = 4
    dose_weights = 5
    limiting_device_angle = 6
    leaves = 7
    MU_number = 8

    ds = pydicom.dcmread(file)[0x300A, 0x00B0].value[beam_sequence_ID][0x300A, 0x0111]
    nb_leaf = 1
    for j in range(len(ds[0][0x300A, 0x011A].value)):
        if ds[0][0x300A, 0x011A].value[j][0x300A, 0x00B8].value == "MLCX":
            nb_leaf = int(len(ds[0][0x300A, 0x011A].value[j][0x300A, 0x011C].value) / 2)

    data_set = [None] * 9
    data_set[jaws_1] = {"jaws_1": []}
    data_set[jaws_2] = {"jaws_2": []}
    data_set[angle] = {"rot_angle": []}
    data_set[dir_angle] = {"dir_angle": []}
    data_set[isocenter] = {"isocenter": []}
    data_set[dose_weights] = {"cumulative weight": []}
    data_set[leaves] = [[{"leaves": []} for _ in range(nb_leaf)] for _ in range(2)]
    data_set[MU_number] = {"MU number": 0}
    data_set[limiting_device_angle] = {"collimation angle": []}
    count = 0

    patient_info = pydicom.dcmread(file)[0x300A, 0x0070]
    number_of_mu = retrieve_value_in_DICOM_RTPlan(
        patient_info.value[0],
        [[("0x300C", "0x0004"), beam_sequence_ID], [("0x300A", "0x0086"), None]],
    )
    data_set[MU_number]["MU number"] = number_of_mu
    for cp in ds.value:
        bool_MLC = False
        bool_jaws = False
        idx_jaws = 0
        idx_MLC = 0
        if (0x300A, 0x011A) in cp:
            for j in range(len(cp[0x300A, 0x011A].value)):
                if cp[0x300A, 0x011A].value[j][0x300A, 0x00B8].value == "MLCX":
                    idx_MLC = j
                    bool_MLC = True

                if cp[0x300A, 0x011A].value[j][0x300A, 0x00B8].value == "ASYMY":
                    idx_jaws = j
                    bool_jaws = True
        if bool_jaws:
            data_set[jaws_1]["jaws_1"].append(
                retrieve_value_in_DICOM_RTPlan(
                    cp, [[("0x300A", "0x011A"), idx_jaws], [("0x300A", "0x011C"), 0]]
                )
            )
            data_set[jaws_2]["jaws_2"].append(
                retrieve_value_in_DICOM_RTPlan(
                    cp, [[("0x300A", "0x011A"), idx_jaws], [("0x300A", "0x011C"), 1]]
                )
            )
        else:
            data_set[jaws_1]["jaws_1"].append(None)
            data_set[jaws_2]["jaws_2"].append(None)

        data_set[angle]["rot_angle"].append(
            retrieve_value_in_DICOM_RTPlan(cp, [[("0x300A", "0x011E"), None]])
        )
        data_set[dir_angle]["dir_angle"].append(
            retrieve_value_in_DICOM_RTPlan(cp, [[("0x300A", "0x011F"), None]])
        )
        data_set[dose_weights]["cumulative weight"].append(
            retrieve_value_in_DICOM_RTPlan(
                cp, [[("0x300C", "0x0050"), 0], [("0x300A", "0x010C"), None]]
            )
        )
        data_set[limiting_device_angle]["collimation angle"].append(
            retrieve_value_in_DICOM_RTPlan(cp, [[("0x300A", "0x0120"), None]])
        )
        data_set[isocenter]["isocenter"].append(ds.value[0]["0x300A", "0x012C"].value)
        if bool_MLC:
            for side_ID in range(2):
                for i in range(nb_leaf):
                    data_set[leaves][side_ID][i]["leaves"].append(
                        retrieve_value_in_DICOM_RTPlan(
                            cp,
                            [
                                [("0x300A", "0x011A"), idx_MLC],
                                [("0x300A", "0x011C"), 80 * side_ID + i],
                            ],
                        )
                    )
        else:
            for side_ID in range(2):
                data_set[leaves][side_ID][0]["leaves"].append(None)

        count += 1
    return data_set



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.userhooks import check_production_cuts
from opengate.tests import utility


def create_simu(nb_threads, paths):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = nb_threads
    sim.random_seed = 987654321
    sim.check_volumes_overlap = False
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq

    # world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # material
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # fake spect head
    waterbox = sim.add_volume("Box", "spect")
    waterbox.size = [55 * cm, 42 * cm, 18 * cm]
    waterbox.material = "G4_AIR"

    # crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = "spect"
    crystal.size = [0.5 * cm, 0.5 * cm, 2 * cm]
    crystal.material = "NaITl"
    crystal.translation = gate.geometry.utility.get_grid_repetition(
        size=[100, 80, 1],
        spacing=[0.5 * cm, 0.5 * cm, 0],
        start=[-25 * cm, -20 * cm, 4 * cm],
    )
    crystal.color = [1, 1, 0, 1]

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 0.01 * mm
    sim.physics_manager.global_production_cuts.electron = 0.01 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    sim.physics_manager.global_production_cuts.proton = 1 * mm

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 4 * cm
    source.position.translation = [0, 0, -15 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 200000 * Bq / sim.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = crystal.name
    hc.output_filename = None  # ""  # paths.output / 'test039_hits.root'
    hc.clear_every = 1
    hc.attributes = [
        "TotalEnergyDeposit",
        "KineticEnergy",
        "PostPosition",
        "PreStepUniqueVolumeID",
        "TrackCreatorProcess",
        "GlobalTime",
        "TrackVolumeName",
        "RunID",
        "ThreadID",
        "TrackID",
    ]

    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.attached_to = crystal.name
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    sc.clear_every = 333
    sc.output_filename = f"test039_singles_{sim.number_of_threads}.root"

    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    cc.attached_to = crystal.name
    cc.input_digi_collection = "Singles"
    cc.clear_every = 10
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
        {
            "name": "spectrum",
            "min": 0 * keV,
            "max": 5000 * keV,
        },  # should be strictly equal to 'Singles'
    ]
    cc.output_filename = "test039_win_e.root"

    # set hook function to dump cuts from G4
    sim.user_hook_after_init = check_production_cuts

    return sim


# go
# sim.run()

# On linux
# valgrind --tool=massif --massif-out-file=./massif_t039_no_cleared.out  python test039_hits_memory_check_MP.py
# valgrind --tool=massif --massif-out-file=./massif_t039_cleared.out  python test039_hits_memory_check_MP.py
# test with clear_every = 1e8 (no-cleared) and with clear_every = 1 (cleared)

"""ms_print arguments: massif_t039_no_cleared.out
--------------------------------------------------------------------------------

    MB
722.9^                                             ##
     |                                             # :::::::@::::::@::::::::::
     |                                         @   # ::: :::@::::::@::       :
     |                                         @:::# ::: :::@::::::@::       :
     |                                         @:::# ::: :::@::::::@::       :
     |                                       ::@:::# ::: :::@::::::@::       :
     |                                       : @:::# ::: :::@::::::@::       :
     |                                      :: @:::# ::: :::@::::::@::       :
     |                              @       :: @:::# ::: :::@::::::@::       :
     |                             @@::::::::: @:::# ::: :::@::::::@::       :
     |                           ::@@::::::::: @:::# ::: :::@::::::@::       :
     |                           ::@@::::::::: @:::# ::: :::@::::::@::       :
     |                         ::::@@::::::::: @:::# ::: :::@::::::@::       :
     |                      :::::::@@::::::::: @:::# ::: :::@::::::@::       :
     |                    :::::::::@@::::::::: @:::# ::: :::@::::::@::       :
     |                  :::::::::::@@::::::::: @:::# ::: :::@::::::@::       :
     |                 ::::::::::::@@::::::::: @:::# ::: :::@::::::@::       :
     |          @:@@:::::::::::::::@@::::::::: @:::# ::: :::@::::::@::       :
     |   ::@:::@@:@@:::::::::::::::@@::::::::: @:::# ::: :::@::::::@::       :
     | ::::@:::@@:@@:::::::::::::::@@::::::::: @:::# ::: :::@::::::@::       :
   0 +----------------------------------------------------------------------->Gi
     0                                                                   143.9
"""

"""ms_print arguments: massif_t039_cleared.out
    MB
154.3^                                                #
     |                :@:::::@::::::::::@:::::::::::@:#::::@::::@::::@@@@@@@@
     |               ::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |             @:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |             @:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |           ::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |         ::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |        @::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |     :::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |    @:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |    @:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |  ::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |  ::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     | :::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
     |@:::@:::@::::@:::@:::::@::::::::::@:::::::::::@:#::::@::::@::::@       :
   0 +----------------------------------------------------------------------->Gi
     0                                                                   144.0
"""


def test_results(sim, paths):
    # Compare stats file
    stats = sim.get_actor("Stats")
    # stats.write(paths.output_ref / 'test039_stats.txt')
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "test039_stats.txt")
    stats.counts.runs = 2  # sim.number_of_threads
    is_ok = utility.assert_stats(stats, stats_ref, 0.05)

    # Compare singles
    print()
    sc = sim.get_actor("Singles")
    gate.exception.warning(f"Check singles")
    ref_file = paths.output_ref / "test039_singles.root"
    hc_file = sc.get_output_path()
    checked_keys = [
        "GlobalTime",
        "TotalEnergyDeposit",
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
    ]
    scalings = [1.0] * len(checked_keys)
    tols = [10.0] * len(checked_keys)
    scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
    tols[checked_keys.index("GlobalTime")] = 0.003
    tols[checked_keys.index("TotalEnergyDeposit")] = 0.001
    tols[checked_keys.index("PostPosition_X")] = 0.2
    tols[checked_keys.index("PostPosition_Y")] = 0.2
    tols[checked_keys.index("PostPosition_Z")] = 0.2
    print(scalings, tols)
    is_ok = (
        utility.compare_root3(
            ref_file,
            hc_file,
            "Singles",
            "Singles",
            checked_keys,
            checked_keys,
            tols,
            scalings,
            scalings,
            paths.output / "test039_singles.png",
            hits_tol=1,
        )
        and is_ok
    )

    return is_ok
    # this is the end, my friend
    # gate.delete_run_manager_if_needed(sim)
    # utility.test_ok(is_ok)
